/*** RelayCard.cpp ***/
// Copyright (C) 2009 by Torsten Knorr
// All rights reserved!
// create-soft@freenet.de
//-----------------------------------------------
 #include "stdafx.h"
 #include "RelayCard.h"
//-----------------------------------------------
// Define timeouts.
 const DWORD CRelayCard::m_WRITE_TIMEOUT 				= 250;
 const DWORD CRelayCard::m_READ_TIMEOUT					= 250;
// Define the realycard protocol.
 const unsigned char CRelayCard::m_NO_OPERATION			= 0;
 const unsigned char CRelayCard::m_NO_OPERATION_OK	= 255;
 const unsigned char CRelayCard::m_SETUP							= 1;
 const unsigned char CRelayCard::m_SETUP_OK					= 254;
 const unsigned char CRelayCard::m_GET_PORT					= 2;
 const unsigned char CRelayCard::m_GET_PORT_OK				= 253;
 const unsigned char CRelayCard::m_SET_PORT					= 3;
 const unsigned char CRelayCard::m_SET_PORT_OK				= 252;
 const unsigned char CRelayCard::m_GET_OPTION				= 4;
 const unsigned char CRelayCard::m_GET_OPTION_OK			= 251;
 const unsigned char CRelayCard::m_SET_OPTION				= 5;
 const unsigned char CRelayCard::m_SET_OPTION_OK			= 250;
 const unsigned char CRelayCard::m_SET_SINGLE					= 6;
 const unsigned char CRelayCard::m_SET_SINGLE_OK			= 249;
 const unsigned char CRelayCard::m_DEL_SINGLE				= 7;
 const unsigned char CRelayCard::m_DEL_SINGLE_OK			= 248;
 const unsigned char CRelayCard::m_TOGGLE						= 8;
 const unsigned char CRelayCard::m_TOGGLE_OK				= 247;
// Define broadcast options.
 const unsigned char CRelayCard::m_BROADCAST_ADDR		= 0;
 const unsigned char CRelayCard::m_NOEXE_NOBLOCK 		= 0;
 const unsigned char CRelayCard::m_EXE_NOBLOCK			= 1;
 const unsigned char CRelayCard::m_NOEXE_BLOCK			= 2;
 const unsigned char CRelayCard::m_EXE_BLOCK					= 3;
 const unsigned char CRelayCard::m_END_OPTIONS				= 3;
// Define bits for the relays.
 const unsigned char CRelayCard::m_RELAY_1						= 1;
 const unsigned char CRelayCard::m_RELAY_2						= 2;
 const unsigned char CRelayCard::m_RELAY_3						= 4;
 const unsigned char CRelayCard::m_RELAY_4						= 8;
 const unsigned char CRelayCard::m_RELAY_5						= 16;
 const unsigned char CRelayCard::m_RELAY_6						= 32;
 const unsigned char CRelayCard::m_RELAY_7						= 64;
 const unsigned char CRelayCard::m_RELAY_8						= 128;
//-----------------------------------------------
 CRelayCard::CRelayCard(void) :
	m_handle_port(NULL),
	m_cards_count(0)
	{
	ZeroMemory(&m_frame_write, sizeof(FRAME));
	ZeroMemory(&m_frame_read, sizeof(FRAME));
	ZeroMemory(&m_frame_list, sizeof(FRAME_LIST));
	ZeroMemory(&m_dcb_settings_old, sizeof(DCB));
	ZeroMemory(&m_dcb_settings_new, sizeof(DCB));
	ZeroMemory(&m_commtimeouts_old, sizeof(COMMTIMEOUTS));
	ZeroMemory(&m_commtimeouts_new, sizeof(COMMTIMEOUTS));
	}
//-----------------------------------------------
 CRelayCard::~CRelayCard(void)
	{
	DeleteFrameList();
	if(!m_handle_port)
		return;
// Configure the port according to the specifications of the DCB structure.
	if(!SetCommState(m_handle_port, &m_dcb_settings_old))
// Could not configure the serial port.
		ErrorMessage(TEXT("CRelayCard::~CRelayCard : SetCommState"));
	if(!SetCommTimeouts(m_handle_port, &m_commtimeouts_old))
// Error setting time-outs.
		ErrorMessage(TEXT("CRelayCard::~CRelayCard : SetCommTimeouts"));
	if(!CloseHandle(m_handle_port))
		ErrorMessage(TEXT("CRelayCard::~CRelayCard : CloseHandle"));
	}
//-----------------------------------------------
 BOOL CRelayCard::Release(void)
	{
	DeleteFrameList();
	if(!m_handle_port)
		return FALSE;
// Configure the port according to the specifications of the DCB structure.
	if(!SetCommState(m_handle_port, &m_dcb_settings_old))
// Could not configure the serial port.
		return ErrorMessage(TEXT("CRelayCard::Release : SetCommState"));
	if(!SetCommTimeouts(m_handle_port, &m_commtimeouts_old))
// Error setting time-outs.
		return ErrorMessage(TEXT("CRelayCard::Release : SetCommTimeouts"));
	if(!CloseHandle(m_handle_port))
		return ErrorMessage(TEXT("CRelayCard::Release : CloseHandle"));
	return TRUE;
	}
//-----------------------------------------------
 BOOL CRelayCard::Init(TCHAR * sz_port)
	{
	m_handle_port = CreateFile(
// pointer to name of the file
		sz_port,
 // access (read-write) mode
		GENERIC_READ | GENERIC_WRITE,
// share mode : exclusive access 
		0,
// pointer to security attributes : default security attributes 
		NULL,
// how to create
		OPEN_EXISTING,
// file attributes
		FILE_ATTRIBUTE_NORMAL | FILE_FLAG_WRITE_THROUGH,
// handle to file with attributes to copy
		NULL
		);
	if(m_handle_port == INVALID_HANDLE_VALUE)
		return ErrorMessage(TEXT("CRelayCard::Init : CreateFile"));
//-----------------------------------------------
// Configure the Serial Port
//-----------------------------------------------
/*** DCBlength ***/
/*
 Before calling the GetCommState function, set this member to the length of the DCB structure.
 Neglecting to do this can cause a failure or return erroneous data.
 Initialize the DCBlength member.
*/
// set size of DCB:
	m_dcb_settings_old.DCBlength = sizeof(DCB);
	m_dcb_settings_new.DCBlength = sizeof(DCB);
//-----------------------------------------------
// Get the default timeouts.
	GetCommTimeouts(m_handle_port, &m_commtimeouts_old);
	CopyMemory(&m_commtimeouts_new, &m_commtimeouts_old, sizeof(COMMTIMEOUTS));
// Get the default port setting information.
	GetCommState (m_handle_port, &m_dcb_settings_old);
	CopyMemory(&m_dcb_settings_new, &m_dcb_settings_old, sizeof(DCB));
// Change the DCB structure settings.
/*** BaudRate ***/
/*
 Specifies the device communication rate.
 Assigns an actual baud rate or an index by specifying a CBR_ constant.
*/
// Current baud 19200:
	m_dcb_settings_new.BaudRate = 19200;
//-----------------------------------------------
/*** fBinary ***/
/*
 Specifies if binary mode is enabled.
 The Win32 API does not support nonbinary mode transfers, so this member must be TRUE. 
 Using FALSE will not work.
*/
// Binary mode, no EOF check:
	m_dcb_settings_new.fBinary = TRUE;
//-----------------------------------------------
/*** fParity ***/
/*
 Specifies if parity checking is enabled.
 If this member is TRUE, parity checking is performed and errors are reported.
*/ 
// Enable parity checking:
	m_dcb_settings_new.fParity = TRUE;
//----------------------------------------------
/*** fOutxCtsFlow ***/
/*
 Turns the CTS flow control on and off.
 To use RTS/CTS flow control, specify TRUE for this member and
 RTS_CONTROL_HANDSHAKE for the fRtsControl member.
*/
// No CTS output flow control:
	//m_dcb_settings_new.fOutxCtsFlow = FALSE;
//-----------------------------------------------
/*** fOutxDsrFlow ***/
/* 
 Turns the DSR flow control on and off. DSR flow control is rarely used.
 A typical port configuration is to set this member to FALSE, while enabling the DTR line.
*/
// No DSR output flow control:
	//m_dcb_settings_new.fOutxDsrFlow = FALSE;
//-----------------------------------------------
/*** fDtrControl ***/
/*
 Specifies the DTR flow control.
 DTR_CONTROL_ENABLE turns on the DTR line during the connection.
 DTR_CONTROL_HANDSHAKE turns on DTR handshaking.
 DTR_CONTROL_DISABLE turns off the DTR line.
 DTR Control Flow Values.
 #define DTR_CONTROL_DISABLE    0x00
 #define DTR_CONTROL_ENABLE     0x01
 #define DTR_CONTROL_HANDSHAKE  0x02
*/
// enable DTR flow control type:
	//m_dcb_settings_new.fDtrControl = DTR_CONTROL_ENABLE;
//-----------------------------------------------
/*** fDsrSensitivity ***/
/*
 Specifies if the communications driver is sensitive to the state of the DSR signal.
 If this member is TRUE, the driver ignores any bytes received,
 unless the DSR modem input line is high.
*/
// no DSR sensitivity:
	m_dcb_settings_new.fDsrSensitivity = FALSE;
//-----------------------------------------------
/*** fTXContinueOnXoff ***/
/*
 Specifies if transmission stops when the input buffer is full and the driver has transmitted
 the XoffChar character. If this member is TRUE, transmission continues after the input
 buffer has come within XoffLim bytes of being full and the driver has transmitted
 the XoffChar character to stop receiving bytes. If this member is FALSE,
 transmission does not continue until the input buffer is within XonLim bytes of being empty
 and the driver has transmitted the XonChar character to resume reception.
*/
// XOFF continues Tx:
	//m_dcb_settings_new.fTXContinueOnXoff = TRUE;
//-----------------------------------------------
/*** fOutX ***/
/*
 Specifies if XON/XOFF flow control is used during transmission.
 If this member is TRUE, transmission stops when the XoffChar character is received
 and starts again when the XonChar character is received.
*/ 
// No XON/XOFF out flow control:
	//m_dcb_settings_new.fOutX = FALSE;
//-----------------------------------------------
/*** fInX ***/
/*
 Specifies if XON/XOFF flow control is used during reception.
 If this member is TRUE, the XoffChar character is sent when the input buffer comes
 within XoffLim bytes of being full, and the XonChar character is sent when the input buffer comes
 within XonLim bytes of being empty.
*/ 
// No XON/XOFF in flow control:
	//m_dcb_settings_new.fInX = FALSE;
//-----------------------------------------------
/*** fErrorChar ***/
/*
 Specifies if bytes received with parity errors are replaced with the character specified by
 the ErrorChar member.
 If this member is TRUE and the fParity member is TRUE, replacement occurs.
*/ 
// Disable error replacement:
	m_dcb_settings_new.fErrorChar = FALSE;
//-----------------------------------------------
/*** fNull ***/
/*
 Specifies if null bytes are discarded.
 If this member is TRUE, null bytes are discarded when received.
*/
// Disable null stripping:
	m_dcb_settings_new.fNull = FALSE;
//-----------------------------------------------
/*** fRtsControl ***/
/*
 Turns the RTS flow control on and off.
 RTS_CONTROL_ENABLE turns on the RTS line during the connection.
 RTS_CONTROL_HANDSHAKE turns on RTS handshaking.
 RTS_CONTROL_DISABLE turns off the RTS line.
 RTS_CONTROL_TOGGLE Specifies that the RTS line is high if bytes are available for transmission.
 After all buffered bytes have been sent, the RTS line will be low.
 RTS Control Flow Values
 #define RTS_CONTROL_DISABLE    0x00
 #define RTS_CONTROL_ENABLE     0x01
 #define RTS_CONTROL_HANDSHAKE  0x02
 #define RTS_CONTROL_TOGGLE     0x03
*/
// enable RTS flow control:
	//m_dcb_settings_new.fRtsControl = RTS_CONTROL_ENABLE;
//-----------------------------------------------
/*** fAbortOnError ***/
/*
 Specifies if read and write operations are terminated if an error occurs.
 If this member is TRUE, the driver terminates all read and write operations with an error status
 if an error occurs. The driver will not accept any further communications operations
 until the application has acknowledged the error by calling the ClearCommError function.
*/ 
// Abort reads/writes on error:
	m_dcb_settings_new.fAbortOnError = TRUE;
//-----------------------------------------------
/*** fDummy2 ***/
/*
 Reserved, do not use.
*/
// No change:
	//dcb_port.fDummy2	= dcb_port.fDummy2;
//-----------------------------------------------
/*** wReserved ***/
/*
 Not used.
*/ 
// Set to zero:
//	m_dcb_settings_new.wReserved = 0;
//-----------------------------------------------
/*** XonLim ***/
/*
 Specifies the minimum number of bytes accepted in the input buffer before the XON character is sent.
*/
// Set to relaycard framesize:
	//m_dcb_settings_new.XonLim = 4;
//-----------------------------------------------
/*** XoffLim ***/
/*
 Specifies the maximum number of bytes accepted in the input buffer before the XOFF character is sent. 
 The maximum number of bytes accepted is calculated by subtracting this value from the size,
 in bytes, of the input buffer.
*/
// Set to relaycard framesize:
	//m_dcb_settings_new.XoffLim = 4;
//-----------------------------------------------
/*** ByteSize ***/
/*
 Specifies the bits per byte transmitted and received.
 Number of bits/byte, 4 - 8
*/
// Set to normal byte size:
	m_dcb_settings_new.ByteSize = 8;
//-----------------------------------------------
/*** Parity ***/
/*
 Specifies the parity scheme. Do not confuse this member with the fParity member,
 which turns parity on or off.
 Because parity is rarely used, this member can usually be NOPARITY.
 0 - 4 = no, odd, even, mark, space
 #define NOPARITY            0
 #define ODDPARITY           1
 #define EVENPARITY          2
 #define MARKPARITY          3
 #define SPACEPARITY         4
*/
// No parity:
	m_dcb_settings_new.Parity = NOPARITY;
//-----------------------------------------------
/*** StopBits ***/
/* 
 0, 1, 2 = 1, 1.5, 2
#define ONESTOPBIT          0
#define ONE5STOPBITS        1
#define TWOSTOPBITS         2
*/
// One stop bit:
	m_dcb_settings_new.StopBits = ONESTOPBIT;
//-----------------------------------------------
/*** XonChar ***/
/*
 Specifies the value of the XON character for both transmission and reception.
*/
// Set to framesize:
	//m_dcb_settings_new.XonChar = 4;
//-----------------------------------------------
/*** XoffChar ***/
/*
 Specifies the value of the XOFF character for both transmission and reception.
*/
// use default:
	//m_dcb_settings_new.XoffChar = dcb_port.XoffChar;
//-----------------------------------------------
/*** ErrorChar ***/
/*
 Specifies the value of the character used to replace bytes received with a parity error.
*/
// use default:
	//m_dcb_settings_new.ErrorChar = dcb_port.ErrorChar;
//-----------------------------------------------
/*** EofChar ***/
/*
 Specifies the value of the character used to signal the end of data.
*/
// use default:
	//m_dcb_settings_new.EofChar = dcb_port.EofChar;
//-----------------------------------------------
/*** EvtChar ***/
/*
 Specifies the value of the character used to signal an event.
*/
// use default:
	//m_dcb_settings_new.EvtChar = dcb_port.EvtChar;
//----------------------------------------------- 
/*** WReserved1 ***/
/*
 Reserved, do not use. 
*/
// No change:
	//m_dcb_settings_new.wReserved1 = dcb_port.wReserved1;
//-----------------------------------------------
// Configure the port according to the specifications of the DCB structure.
	if(!SetCommState (m_handle_port, &m_dcb_settings_new))
// Could not configure the serial port.
		return ErrorMessage(TEXT("CRelayCard::Init : SetCommState"));
// Set communication timeouts.
/*** ReadIntervalTimeout ***/
// Specifies the maximum acceptable time, in milliseconds,
// to elapse between the arrival of two characters on the communication line.
// During a ReadFile operation, the time period begins when the first character is received.
// If the interval between the arrival of any two characters exceeds this amount,
// the ReadFile operation is completed and any buffered data is returned.
// A value of zero indicates that interval time-outs are not used. 
	m_commtimeouts_new.ReadIntervalTimeout = 50;
/*** ReadTotalTimeoutConstant ***/
// Specifies the constant, in milliseconds, used to calculate the total time-out period for read operations.
// For each read operation, this value is added to the product of the ReadTotalTimeoutMultiplier 
// member and the requested number of bytes. 
// A value of zero for both the ReadTotalTimeoutMultiplier and ReadTotalTimeoutConstant 
// members indicates that total time-outs are not used for read operations. 
	m_commtimeouts_new.ReadTotalTimeoutConstant = m_READ_TIMEOUT;
/*** ReadTotalTimeoutMultiplier ***/
// Specifies the multiplier, in milliseconds, used to calculate the total time-out period for read operations.
//  For each read operation, this value is multiplied by the requested number of bytes to be read.
	m_commtimeouts_new.ReadTotalTimeoutMultiplier = 50;
/*** WriteTotalTimeoutConstant ***/
// Specifies the constant, in milliseconds, used to calculate the total time-out period for write operations.
// For each write operation, this value is added to the product of the WriteTotalTimeoutMultiplier
//  member and the number of bytes to be written. 
// A value of zero for both the WriteTotalTimeoutMultiplier and WriteTotalTimeoutConstant
// members indicates that total time-outs are not used for write operations.
	m_commtimeouts_new.WriteTotalTimeoutConstant = m_WRITE_TIMEOUT;
/*** WriteTotalTimeoutMultiplier ***/
// Specifies the multiplier, in milliseconds, used to calculate the total time-out period for write operations.
// For each write operation, this value is multiplied by the number of bytes to be written. 
	m_commtimeouts_new.WriteTotalTimeoutMultiplier = 50;
	if(!SetCommTimeouts(m_handle_port, &m_commtimeouts_new))
		return ErrorMessage(TEXT("CRelayCard::Init : SetCommTimeouts"));
// Set serial communication mask.
	if(!SetCommMask(m_handle_port, EV_RXCHAR))
// Error setting communications event mask
		ErrorMessage(TEXT("CRelayCard::Init : SetCommMask"));
	m_frame_write.command = m_SETUP;
	m_frame_write.address	= 1;
	m_frame_write.data			= 0;
	m_frame_write.checksum	= 0;
	if(!WriteFrame())
		return ErrorMessage(TEXT("CRelayCard::Init : WriteCommand"));
	return TRUE;
	}
//-----------------------------------------------
 BOOL CRelayCard::SendData(void)
	{
	if(8 < m_frame_write.command)
		return FALSE;
	if(m_cards_count < m_frame_write.address)
		return FALSE;
	m_frame_write.checksum = m_frame_write.command ^ m_frame_write.address ^ m_frame_write.data;
	return WriteFrame();
	}
//-----------------------------------------------
 BOOL CRelayCard::WriteFrame(void)
	{
	DWORD dw_written = 0;
	DeleteFrameList();
	m_frame_list.frame.command	= m_frame_write.command;
	m_frame_list.frame.address		= m_frame_write.address;
	m_frame_list.frame.data			= m_frame_write.data;
	m_frame_list.frame.checksum	= m_frame_write.checksum;
	if(!WriteFile(m_handle_port, &m_frame_write, 4, &dw_written, NULL))
		return ErrorMessage(TEXT("CRelayCard::WriteCommand : WriteFile"));
	else
		{
// Write operation completed successfully.
// Is it a successfully write?
		if(4 != dw_written)
			return ErrorMessage(TEXT("CRealyCard::WriteCommand : WriteFile : writes more or less as 4 bytes"));
		else
			return ReadResponse();
		}
	}
//-----------------------------------------------
 BOOL CRelayCard::ReadResponse(void)
	{
	DWORD dw_read = 0;
// Issue read operation.
	while(ReadFile(m_handle_port, &m_frame_read, 4, &dw_read, NULL))
		{
		if(0 == dw_read)
			break;
// Is it a successful read?
		if(4 != dw_read)
			return ErrorMessage(TEXT("CRelayCard::ReadResponse : ReadFile : read more or less as 4 bytes"));
		AddNewFrame(
			m_frame_read.command,
			m_frame_read.address,
			m_frame_read.data,
			m_frame_read.checksum
			);
		if(!CheckResponse())
			return FALSE;
		}
	return TRUE;
	}
//-----------------------------------------------
 BOOL CRelayCard::CheckResponse(void)
	{
#ifdef _DEBUG
	TCHAR sz_buffer[256];
	wsprintf(
		sz_buffer,
		TEXT("frame write:\t%u\t%u\t%u\t%u\n")
		TEXT("frame read:\t%u\t%u\t%u\t%u"),
		m_frame_write.command,
		m_frame_write.address,
		m_frame_write.data,
		m_frame_write.checksum,
		m_frame_read.command,
		m_frame_read.address,
		m_frame_read.data,
		m_frame_read.checksum
		);
	MessageBox(NULL, sz_buffer, TEXT("CRelayCard::CheckResponse"), MB_OK);
#endif
	if(!
		m_frame_read.checksum
		==
		(m_frame_read.command ^ m_frame_read.address ^ m_frame_read.data)
		)
		return ErrorMessage(TEXT("CRelayCard::CheckResponse : Checksumerror"));
	switch(m_frame_write.command)
		{
		case m_NO_OPERATION:
			if(!(
				(m_NO_OPERATION_OK == m_frame_read.command)
				||
				((m_BROADCAST_ADDR == m_frame_write.address)
				&&
				(m_NO_OPERATION == m_frame_read.command))
				))
				return ErrorMessage(TEXT("CRelayCard::CheckResponse : ERROR - NO_OPERATION"));
			break;
		case m_SETUP:
			if(!(
				(m_SETUP_OK == m_frame_read.command)
				||
				(m_SETUP == m_frame_read.command)
				))
				return ErrorMessage(TEXT("CRelayCard::CheckResponse : ERROR - SETUP"));
			m_cards_count = m_frame_read.address - 1;
			break;
		case m_GET_PORT:
			if(!(
				(m_GET_PORT_OK == m_frame_read.command)
				||
// Is it a broadcast?
				((m_BROADCAST_ADDR == m_frame_write.address)
				&&
				((m_GET_PORT == m_frame_read.command)
				||
// Is broadcast blocked?
				(m_NO_OPERATION == m_frame_read.command)
				||
				(m_NO_OPERATION_OK == m_frame_read.command)))
				))
				return ErrorMessage(TEXT("CRelayCard::CheckResponse : ERROR - GET_PORT"));
			break;
		case m_SET_PORT:
			if(!(
				(m_SET_PORT_OK == m_frame_read.command)
				||
				((m_BROADCAST_ADDR == m_frame_write.address)
				&&
				((m_SET_PORT == m_frame_read.command)
				||
				(m_NO_OPERATION == m_frame_read.command)
				||
				(m_NO_OPERATION_OK == m_frame_read.command))) 
				))
				return ErrorMessage(TEXT("CRelayCard::CheckResponse : ERROR - SET_PORT"));
			break;
		case m_GET_OPTION:
			if(!(
				(m_GET_OPTION_OK == m_frame_read.command)
				||
				((m_BROADCAST_ADDR == m_frame_write.address)
				&&
				((m_GET_OPTION == m_frame_read.command)
				||
				(m_NO_OPERATION == m_frame_read.command)
				||
				(m_NO_OPERATION_OK == m_frame_read.command)))
				))
				return ErrorMessage(TEXT("CRelayCard::CheckResponse : ERROR - GET_OPTION"));
			break;
		case m_SET_OPTION:
			if(!(
				(m_SET_OPTION_OK == m_frame_read.command)
				||
				((m_BROADCAST_ADDR == m_frame_write.address)
				&&
				((m_SET_OPTION == m_frame_read.command)
				||
				(m_NO_OPERATION == m_frame_read.command)
				||
				(m_NO_OPERATION_OK == m_frame_read.command)))
				))
				return ErrorMessage(TEXT("CRelayCard::CheckResponse : ERROR - SET_OPTION"));
			break;
		case m_SET_SINGLE:
			if(!(
				(m_SET_SINGLE_OK == m_frame_read.command)
				||
				((m_BROADCAST_ADDR == m_frame_write.address)
				&&
				((m_SET_SINGLE == m_frame_read.command)
				||
				(m_NO_OPERATION == m_frame_read.command)
				||
				(m_NO_OPERATION_OK == m_frame_read.command)))
				))
				return ErrorMessage(TEXT("CRelayCard::CheckResponse : ERROR - SET_SINGLE"));
			break;
		case m_DEL_SINGLE:
			if(!(
				(m_DEL_SINGLE_OK == m_frame_read.command)
				||
				((m_BROADCAST_ADDR == m_frame_write.address)
				&&
				((m_DEL_SINGLE == m_frame_read.command)
				||
				(m_NO_OPERATION == m_frame_read.command)
				||
				(m_NO_OPERATION_OK == m_frame_read.command)))
				))
				return ErrorMessage(TEXT("CRelayCard::CheckResponse : ERROR - DEL_SINGLE"));
			break;
		case m_TOGGLE:
			if(!(
				(m_TOGGLE_OK == m_frame_read.command)
				||
				((m_BROADCAST_ADDR == m_frame_read.address)
				&&
				((m_TOGGLE == m_frame_read.command)
				||
				(m_NO_OPERATION == m_frame_read.command)
				||
				(m_NO_OPERATION_OK == m_frame_read.command)))
				))
				return ErrorMessage(TEXT("CRelayCard::CheckResponse : ERROR - TOGGLE"));
			break;
		default:
			return ErrorMessage(TEXT("CRelayCard::CheckResponse : ERROR - undefined command"));
			break;  
		}
	return TRUE;
	}
//-----------------------------------------------
 BOOL CRelayCard::ErrorMessage(TCHAR * sz_message)
	{
// format for wsprintf
	TCHAR * sz_format = "ERROR: %d: %s.\n\r%s\r\n";
	TCHAR * sz_error_string;
	TCHAR * sz_output_string;
	DWORD dw_error_code = GetLastError();
	DWORD dw_string_length = 0;
// Get error string from system
	dw_string_length = FormatMessage(
		FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_ALLOCATE_BUFFER | 80 ,
		NULL,
		dw_error_code,
		//MAKELANGID(LANG_ENGLISH, SUBLANG_ENGLISH_US),
		//MAKELANGID(LANG_GERMAN, SUBLANG_GERMAN),
// Default language
		0,
		(LPTSTR)&sz_error_string,
		0,
		NULL
		);
	if(dw_string_length == 0)
		sz_error_string = (TCHAR*)LocalAlloc(LPTR, 1);
// Allocate buffer for error string from system, passed in string
// and extra stuff from the szFormat string.
	sz_output_string = (TCHAR*)LocalAlloc(LPTR, strlen(sz_message) + dw_string_length + 30);
// If no final buffer, then can't format error.
	if(NULL == sz_output_string)
		MessageBox(NULL, TEXT("Cannot properly report error."), TEXT("Fatal Error"), MB_ICONERROR);
    else
		{	
		wsprintf(sz_output_string, sz_format, dw_error_code, sz_message, sz_error_string);
		OutputDebugString(sz_output_string);
		MessageBox(NULL, sz_output_string, TEXT("MyErrorMessage"), MB_ICONERROR);
// Free final buffer.
		LocalFree(sz_output_string);
		}
// Free extended string buffer.
	LocalFree(sz_error_string);
	return FALSE;
	}
//-----------------------------------------------
 unsigned char CRelayCard::GetCardsCount(void)
	{
	return m_cards_count;
	}
//-----------------------------------------------
 unsigned char CRelayCard::GetWriteCommand(void)
	{
	return m_frame_write.command;
	}
//-----------------------------------------------
 void CRelayCard::SetWriteCommand(unsigned char uc_command)
	{
	if(m_TOGGLE < uc_command)
		{
		m_frame_write.command = m_NO_OPERATION;
		ErrorMessage(TEXT("CRelayCard::SetWriteCommand : invalid command"));
		}
	else
		m_frame_write.command = uc_command;
	}
//-----------------------------------------------
 unsigned char CRelayCard::GetWriteAddress(void)
	{
	return m_frame_write.address;
	}
//-----------------------------------------------
  void CRelayCard::SetWriteAddress(unsigned char uc_address)
	{
	if(m_cards_count < uc_address)
		ErrorMessage(TEXT("CRelayCard::SetWriteAddress : invalid card address"));
	else
		m_frame_write.address = uc_address;
	}
//-----------------------------------------------
 unsigned char CRelayCard::GetWriteData(void)
	{
	return m_frame_write.data;
	}
//-----------------------------------------------
 void CRelayCard::SetWriteData(unsigned char uc_data)
	{
	m_frame_write.data = uc_data;
	}
//-----------------------------------------------
/*** SET_SINGLE ***/
 BOOL CRelayCard::RelayOn(unsigned char uc_card, unsigned char uc_relay)
	{
	if(m_cards_count < uc_card)
		return ErrorMessage(TEXT("CRelayCard::RelayOn : invalid card address"));
	if((0 == uc_relay) || (8 < uc_relay))
		return ErrorMessage(TEXT("CRelayCard::RelayOn : invalid relay number"));
	unsigned char uc_bits = (unsigned char)pow(2, --uc_relay);
	m_frame_write.command	= m_SET_SINGLE;
	m_frame_write.address	= uc_card;
	m_frame_write.data			= uc_bits;
	m_frame_write.checksum	= m_SET_SINGLE ^ uc_card ^ uc_bits;
	return WriteFrame();
	}
//-----------------------------------------------
/*** DEL_SINGLE ***/
 BOOL CRelayCard::RelayOff(unsigned char uc_card, unsigned char uc_relay)
	{
	if(m_cards_count < uc_card)
		return ErrorMessage(TEXT("CRelayCard::RelayOff : invalid card address"));
	if((0 == uc_relay) || (8 < uc_relay))
		return ErrorMessage(TEXT("CRelayCard::RelayOff : invalid relay number"));
	unsigned char uc_bits = (unsigned char)pow(2, --uc_relay);
	m_frame_write.command	= m_DEL_SINGLE;
	m_frame_write.address	= uc_card;
	m_frame_write.data			= uc_bits;
	m_frame_write.checksum	= m_DEL_SINGLE ^ uc_card ^ uc_bits;
	return WriteFrame();
	}
//-----------------------------------------------
 BOOL CRelayCard::AllRelaysOn(void)
	{
	return RelayCardOn(0);
	}
//-----------------------------------------------
/*** SET_PORT ***/
 BOOL CRelayCard::RelayCardOn(unsigned char uc_card)
	{
	if(m_cards_count < uc_card)
		return ErrorMessage(TEXT("CRelayCard::RelayCardOn : invalid card address"));
	m_frame_write.command	= m_SET_PORT;
	m_frame_write.address	= uc_card;
	m_frame_write.data			= 255;
	m_frame_write.checksum	= m_SET_PORT ^ uc_card ^ 255;
	return WriteFrame();
	}
//-----------------------------------------------
 BOOL CRelayCard::AllRelaysOff(void)
	{
	return RelayCardOff(0);
	}
//-----------------------------------------------
/*** SET_PORT ***/
 BOOL CRelayCard::RelayCardOff(unsigned char uc_card)
	{
	if(m_cards_count < uc_card)
		return ErrorMessage(TEXT("CRelayCard::RelayCardOff : invalid card address"));
	m_frame_write.command	= m_SET_PORT;
	m_frame_write.address	= uc_card;
	m_frame_write.data			= 0;
	m_frame_write.checksum	= m_SET_PORT ^ uc_card;
	return WriteFrame();
	}
//-----------------------------------------------
 BOOL CRelayCard::AllRelaysToggle(unsigned char uc_status_byte)
	{
	return RelayCardToggle(0, uc_status_byte);
	}
//-----------------------------------------------
/*** TOGGLE ***/
 BOOL CRelayCard::RelayCardToggle(unsigned char uc_card, unsigned char uc_status_byte)
	{
	if(m_cards_count < uc_card)
		return ErrorMessage(TEXT("CRelayCard::RelayCardToggle : invalid card address"));
	m_frame_write.command	= m_TOGGLE;
	m_frame_write.address	= uc_card;
	m_frame_write.data			= uc_status_byte;
	m_frame_write.checksum	= m_TOGGLE ^ uc_card ^ uc_status_byte;
	return WriteFrame();
	}
//-----------------------------------------------
/*** SET_PORT ***/
 BOOL CRelayCard::SetRelays(unsigned char uc_card, unsigned char uc_status_byte)
	{
	if(m_cards_count < uc_card)
		return ErrorMessage(TEXT("CRelayCard::SetRelays : invalid card address"));
	m_frame_write.command	= m_SET_PORT;
	m_frame_write.address	= uc_card;
	m_frame_write.data			= uc_status_byte;
	m_frame_write.checksum	= m_SET_PORT ^ uc_card ^ uc_status_byte;
	return WriteFrame();
	}
//-----------------------------------------------
/*** GET_PORT ***/
 unsigned char CRelayCard::GetStatusByte(unsigned char uc_card)
	{
// No broadcast!
	if(0 == uc_card)
		return ErrorMessage(TEXT("CRelayCard::GetStatusByte : no broadcast"));
	if(m_cards_count < uc_card)
		return ErrorMessage(TEXT("CRelayCard::GetStatusByte : invalid card address"));
	m_frame_write.command	= m_GET_PORT;
	m_frame_write.address	= uc_card;
	m_frame_write.data			= 0;
	m_frame_write.checksum	= m_GET_PORT ^ uc_card;
	WriteFrame();
	return m_frame_read.data;
	}
//-----------------------------------------------
/*** SET_OPTION ***/
 BOOL CRelayCard::SetOptions(unsigned char uc_card, unsigned char uc_options)
	{
	if(m_cards_count < uc_card)
		return ErrorMessage(TEXT("CRelayCard::SetOptions : invalid card address"));
	if(m_END_OPTIONS < uc_options)
		return ErrorMessage(TEXT("CRelayCard::SetOptions : invalid options"));
	m_frame_write.command	= m_SET_OPTION;
	m_frame_write.address	= uc_card;
	m_frame_write.data			= uc_options;
	m_frame_write.checksum	= m_SET_OPTION ^ uc_card ^ uc_options;
	return WriteFrame();
	}
//-----------------------------------------------
/*** GET_OPTION ***/
 unsigned char CRelayCard::GetOptions(unsigned char uc_card)
	{
// No broadcast!
	if(0 == uc_card)
		return ErrorMessage(TEXT("CRelayCard::GetOptions : no broadcast"));
	if(m_cards_count < uc_card)
		return ErrorMessage(TEXT("CRelayCard::GetOptions : invalid card address"));
	m_frame_write.command	= m_GET_OPTION;
	m_frame_write.address	= uc_card;
	m_frame_write.data			= 0;
	m_frame_write.checksum	= m_GET_OPTION ^ uc_card;
	WriteFrame();
	return m_frame_read.data;
	}
//-----------------------------------------------
 PFRAME_LIST CRelayCard::GetFrameList(void)
	{
	return &m_frame_list;
	}
//-----------------------------------------------
 void CRelayCard::AddNewFrame(
	unsigned char uc_command,
	unsigned char uc_address,
	unsigned char uc_data,
	unsigned char uc_checksum
	)
	{
	PFRAME_LIST p_frame_list = new FRAME_LIST;
	p_frame_list->frame.command	= uc_command;
	p_frame_list->frame.address		= uc_address;
	p_frame_list->frame.data			= uc_data;
	p_frame_list->frame.checksum	= uc_checksum;
	p_frame_list->p_next_frame		= NULL;
	LastFrame()->p_next_frame		= p_frame_list;
	}
//-----------------------------------------------
 PFRAME_LIST CRelayCard::LastFrame(void)
	{
	PFRAME_LIST p_frame_list = &m_frame_list;
	while(NULL != p_frame_list->p_next_frame)
		p_frame_list = p_frame_list->p_next_frame;
	return p_frame_list;
	}
//-----------------------------------------------
 void CRelayCard::DeleteFrameList(void)
	{
	while(NULL != m_frame_list.p_next_frame)
		{
		PFRAME_LIST p_frame_list = &m_frame_list;
		while(LastFrame() != p_frame_list->p_next_frame)
			p_frame_list = p_frame_list->p_next_frame;
		delete LastFrame();
		p_frame_list->p_next_frame = NULL;
		}
	}
//-----------------------------------------------