/*** RelayCard.h ***/
// Copyright (C) 2009 by Torsten Knorr
// All rights reserved!
// create-soft@freenet.de
//-----------------------------------------------
 #pragma once
 #include <windows.h>
 #include <math.h>
//-----------------------------------------------
 typedef struct
	{
	unsigned char command;
	unsigned char address;
	unsigned char data;
	unsigned char checksum;
	} FRAME, * PFRAME;
//-----------------------------------------------
 typedef struct type_FRAME_LIST
	{
	FRAME frame;
	type_FRAME_LIST * p_next_frame;
	} FRAME_LIST, * PFRAME_LIST;
//-----------------------------------------------
 class CRelayCard
	{
	public:
		CRelayCard(void);
		~CRelayCard(void);
		BOOL Release(void);
		BOOL Init(TCHAR * sz_port);
		BOOL RelayOn(unsigned char uc_card, unsigned char uc_relay);
		BOOL RelayOff(unsigned char uc_card, unsigned char uc_relay);
		BOOL AllRelaysOn(void);
		BOOL AllRelaysOff(void);
		BOOL AllRelaysToggle(unsigned char uc_status_byte);
		BOOL RelayCardOn(unsigned char uc_card);
		BOOL RelayCardOff(unsigned char uc_card);
		BOOL RelayCardToggle(unsigned char uc_card, unsigned char uc_status_byte);
		BOOL SetOptions(unsigned char uc_card, unsigned char uc_optionen);
		unsigned char GetOptions(unsigned char uc_card);
		BOOL SetRelays(unsigned char uc_card, unsigned char uc_status_byte);
		unsigned char GetStatusByte(unsigned char uc_card);
		unsigned char GetCardsCount(void);
// Declare bits for the relays.
		static const unsigned char m_RELAY_1;
		static const unsigned char m_RELAY_2;
		static const unsigned char m_RELAY_3;
		static const unsigned char m_RELAY_4;
		static const unsigned char m_RELAY_5;
		static const unsigned char m_RELAY_6;
		static const unsigned char m_RELAY_7;
		static const unsigned char m_RELAY_8;
	protected:
		unsigned char GetWriteCommand(void);
		void SetWriteCommand(unsigned char uc_command);
		unsigned char GetWriteAddress(void);
		void SetWriteAddress(unsigned char uc_address);
		unsigned char GetWriteData(void);
		void SetWriteData(unsigned char uc_data);
		BOOL SendData(void);
		PFRAME_LIST GetFrameList(void);
// Declare timeouts.
		static const DWORD m_WRITE_TIMEOUT;
		static const DWORD m_READ_TIMEOUT;
// Declare the relaycard protocol.
		static const unsigned char m_NO_OPERATION;
		static const unsigned char m_NO_OPERATION_OK;
		static const unsigned char m_SETUP;
		static const unsigned char m_SETUP_OK;
		static const unsigned char m_GET_PORT;
		static const unsigned char m_GET_PORT_OK;
		static const unsigned char m_SET_PORT;
		static const unsigned char m_SET_PORT_OK;
		static const unsigned char m_GET_OPTION;
		static const unsigned char m_GET_OPTION_OK;
		static const unsigned char m_SET_OPTION;
		static const unsigned char m_SET_OPTION_OK;
		static const unsigned char m_SET_SINGLE;
		static const unsigned char m_SET_SINGLE_OK;
		static const unsigned char m_DEL_SINGLE;
		static const unsigned char m_DEL_SINGLE_OK;
		static const unsigned char m_TOGGLE;
		static const unsigned char m_TOGGLE_OK;
// Declare broadcast options.
		static const unsigned char m_BROADCAST_ADDR;
		static const unsigned char m_NOEXE_NOBLOCK;
		static const unsigned char m_EXE_NOBLOCK;
		static const unsigned char m_NOEXE_BLOCK;
		static const unsigned char m_EXE_BLOCK;
		static const unsigned char m_END_OPTIONS;
	private:
		BOOL WriteFrame(void);
		BOOL ReadResponse(void);
		BOOL CheckResponse(void);
		BOOL ErrorMessage(TCHAR * sz_message);
		HANDLE		m_handle_port;
		unsigned char	m_cards_count;
		FRAME			m_frame_write;
		FRAME			m_frame_read;
		DCB m_dcb_settings_old;
		DCB m_dcb_settings_new;
		COMMTIMEOUTS	 m_commtimeouts_old;
		COMMTIMEOUTS	 m_commtimeouts_new;
		FRAME_LIST	m_frame_list;
		PFRAME_LIST	LastFrame(void);
		void	AddNewFrame(
			unsigned char uc_command,
			unsigned char uc_address,
			unsigned char uc_data,
			unsigned char uc_checksum
			);
		void DeleteFrameList(void);
	};
//-----------------------------------------------